import cutagi

from pytagi.nn.base_layer import BaseLayer


class AvgPool2d(BaseLayer):
    """Average Pooling layer"""

    def __init__(
        self,
        kernel_size: int,
        stride: int = -1,
        padding: int = 0,
        padding_type: int = 0,
    ):
        super().__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.padding_type = padding_type

        self._cpp_backend = cutagi.AvgPool2d(
            kernel_size, stride, padding, padding_type
        )

    def get_layer_info(self) -> str:
        return self._cpp_backend.get_layer_info()

    def get_layer_name(self) -> str:
        return self._cpp_backend.get_layer_name()


class MaxPool2d(BaseLayer):
    """Max Pooling layer"""

    def __init__(
        self,
        kernel_size: int,
        stride: int = 1,
        padding: int = 0,
        padding_type: int = 0,
    ):
        super().__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.padding_type = padding_type

        self._cpp_backend = cutagi.MaxPool2d(
            kernel_size, stride, padding, padding_type
        )

    def get_layer_info(self) -> str:
        return self._cpp_backend.get_layer_info()

    def get_layer_name(self) -> str:
        return self._cpp_backend.get_layer_name()



from typing import List, Optional

import cutagi


class BaseHiddenStates:
    def __init__(
        self, size: Optional[int] = None, block_size: Optional[int] = None
    ):
        if size is not None and block_size is not None:
            self._cpp_backend = cutagi.BaseHiddenStates(size, block_size)
        else:
            self._cpp_backend = cutagi.BaseHiddenStates()

    @property
    def mu_a(self) -> List[float]:
        return self._cpp_backend.mu_a

    @mu_a.setter
    def mu_a(self, value: List[float]):
        self._cpp_backend.mu_a = value

    @property
    def var_a(self) -> List[float]:
        return self._cpp_backend.var_a

    @var_a.setter
    def var_a(self, value: List[float]):
        self._cpp_backend.var_a = value

    @property
    def jcb(self) -> List[float]:
        return self._cpp_backend.jcb

    @jcb.setter
    def jcb(self, value: List[float]):
        self._cpp_backend.jcb = value

    @property
    def size(self) -> int:
        return self._cpp_backend.size

    @property
    def block_size(self) -> int:
        return self._cpp_backend.block_size

    @property
    def actual_size(self) -> int:
        return self._cpp_backend.actual_size

    def set_input_x(
        self, mu_x: List[float], var_x: List[float], block_size: int
    ):
        self._cpp_backend.set_input_x(mu_x, var_x, block_size)

    def get_name(self) -> str:
        return self._cpp_backend.get_name()

    def set_size(self, new_size: int, new_block_size: int) -> str:
        self._cpp_backend.set_size(new_size, new_block_size)


class BaseDeltaStates:
    def __init__(
        self, size: Optional[int] = None, block_size: Optional[int] = None
    ):
        if size is not None and block_size is not None:
            self._cpp_backend = cutagi.BaseDeltaStates(size, block_size)
        else:
            self._cpp_backend = cutagi.BaseDeltaStates()

    @property
    def delta_mu(self) -> List[float]:
        return self._cpp_backend.delta_mu

    @delta_mu.setter
    def delta_mu(self, value: List[float]):
        self._cpp_backend.delta_mu = value

    @property
    def delta_var(self) -> List[float]:
        return self._cpp_backend.delta_var

    @delta_var.setter
    def delta_var(self, value: List[float]):
        self._cpp_backend.delta_var = value

    @property
    def size(self) -> int:
        return self._cpp_backend.size

    @property
    def block_size(self) -> int:
        return self._cpp_backend.block_size

    @property
    def actual_size(self) -> int:
        return self._cpp_backend.actual_size

    def get_name(self) -> str:
        return self._cpp_backend.get_name()

    def reset_zeros(self) -> None:
        """Reset all delta_mu and delta_var to zeros"""
        self._cpp_backend.reset_zeros()

    def copy_from(self, source: "BaseDeltaStates", num_data: int = -1) -> None:
        """Copy values of delta_mu and delta_var from delta states"""
        self._cpp_backend.copy_from(source, num_data)

    def set_size(self, new_size: int, new_block_size: int) -> str:
        self._cpp_backend.set_size(new_size, new_block_size)


class HRCSoftmax:
    """Hierarchical softmax wrapper from the CPP backend. Further details can be
    found here https://building-babylon.net/2017/08/01/hierarchical-softmax

    Attributes:
        obs: A fictive observation \in [-1, 1]
        idx: Indices assigned to each label
        num_obs: Number of indices for each label
        len: Length of an observation e.g 10 labels -> len(obs) = 11
    """

    def __init__(self) -> None:
        self._cpp_backend = cutagi.HRCSoftmax()

    @property
    def obs(self) -> List[float]:
        return self._cpp_backend.obs

    @obs.setter
    def obs(self, value: List[float]):
        self._cpp_backend.obs = value

    @property
    def idx(self) -> List[int]:
        return self._cpp_backend.idx

    @idx.setter
    def idx(self, value: List[int]):
        self._cpp_backend.idx = value

    @property
    def num_obs(self) -> int:
        return self._cpp_backend.num_obs

    @num_obs.setter
    def num_obs(self, value: int):
        self._cpp_backend.num_obs = value

    @property
    def len(self) -> int:
        return self._cpp_backend.len

    @len.setter
    def len(self, value: int):
        self._cpp_backend.len = value



from typing import List

import cutagi

from pytagi.nn.base_layer import BaseLayer


class BatchNorm2d(BaseLayer):
    """Batch normalization"""

    def __init__(
        self,
        num_features: int,
        eps: float = 1e-5,
        momentum: float = 0.9,
        bias: bool = True,
        gain_weight: float = 1.0,
        gain_bias: float = 1.0,
    ):
        self.num_features = num_features
        self.eps = eps
        self.momentum = momentum
        self.is_bias = bias
        self._cpp_backend = cutagi.BatchNorm2d(
            num_features, eps, momentum, bias, gain_weight, gain_bias
        )

    def get_layer_info(self) -> str:
        return self._cpp_backend.get_layer_info()

    def get_layer_name(self) -> str:
        return self._cpp_backend.get_layer_name()

    def init_weight_bias(self):
        self._cpp_backend.init_weight_bias()



from typing import List, Tuple

import cutagi
import numpy as np

from pytagi.nn.base_layer import BaseLayer
from pytagi.nn.data_struct import BaseDeltaStates, BaseHiddenStates


class Sequential:
    """Adding neural networks in a sequence mode"""

    def __init__(self, *layers: BaseLayer):
        """
        Initialize the Sequential model with the given layers.
        Args:
            layers: A variable number of layers (instances of BaseLayer or derived classes).
        """
        backend_layers = [layer._cpp_backend for layer in layers]
        self._cpp_backend = cutagi.Sequential(backend_layers)

    def __call__(
        self, mu_x: np.ndarray, var_x: np.ndarray = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        return self.forward(mu_x, var_x)

    @property
    def layers(self) -> List[BaseLayer]:
        """Get the layers of the model."""
        return self._cpp_backend.layers

    @property
    def output_z_buffer(self) -> BaseHiddenStates:
        """Get the output hidden states"""
        return self._cpp_backend.output_z_buffer

    @output_z_buffer.setter
    def output_z_buffer(self, value: BaseHiddenStates):
        """Set the output hidden states."""
        self._cpp_backend.output_z_buffer = value

    @property
    def input_delta_z_buffer(self) -> BaseDeltaStates:
        """Get the delta hidden states"""
        return self._cpp_backend.input_delta_z_buffer

    @input_delta_z_buffer.setter
    def input_delta_z_buffer(self, value: BaseDeltaStates):
        """Set the delta hidden states."""
        self._cpp_backend.input_delta_z_buffer = value

    @property
    def output_delta_z_buffer(self) -> BaseDeltaStates:
        """Get the delta hidden states"""
        return self._cpp_backend.output_delta_z_buffer

    @output_delta_z_buffer.setter
    def output_delta_z_buffer(self, value: BaseDeltaStates):
        """Set the delta hidden states."""
        self._cpp_backend.output_delta_z_buffer = value

    @property
    def z_buffer_size(self) -> int:
        """Get the z_buffer_size."""
        return self._cpp_backend.z_buffer_size

    @z_buffer_size.setter
    def z_buffer_size(self, value: int):
        """Set the z_buffer_size."""
        self._cpp_backend.z_buffer_size = value

    @property
    def z_buffer_block_size(self) -> int:
        """Get the z_buffer_block_size."""
        return self._cpp_backend.z_buffer_block_size

    @z_buffer_block_size.setter
    def z_buffer_block_size(self, value: int):
        """Set the z_buffer_block_size."""
        self._cpp_backend.z_buffer_block_size = value

    @property
    def device(self) -> str:
        """Get the device"""
        return self._cpp_backend.device

    @device.setter
    def device(self, value: str):
        """Set the sevice"""
        self._cpp_backend.device = value

    @property
    def input_state_update(self) -> bool:
        """Get the device"""
        return self._cpp_backend.input_state_update

    @input_state_update.setter
    def input_state_update(self, value: bool):
        """Set the sevice"""
        self._cpp_backend.input_state_update = value

    @property
    def num_samples(self) -> int:
        """Get the num_samples."""
        return self._cpp_backend.num_samples

    @num_samples.setter
    def num_samples(self, value: int):
        """Set the num_samples."""
        self._cpp_backend.num_samples = value

    def to_device(self, device: str):
        """Move the model to a specific device."""
        self._cpp_backend.to_device(device)

    def params_to_device(self):
        """Move the model parameters to a specific cuda device."""
        self._cpp_backend.params_to_device()

    def params_to_host(self):
        """Move the model parameters from cuda device to the host."""
        self._cpp_backend.params_to_host()

    def set_threads(self, num_threads: int):
        """Set the number of threads to use."""
        self._cpp_backend.set_threads(num_threads)

    def train(self):
        """Set the number of threads to use."""
        self._cpp_backend.train()

    def eval(self):
        """Set the number of threads to use."""
        self._cpp_backend.eval()

    def forward(
        self, mu_x: np.ndarray, var_x: np.ndarray = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Perform a forward pass."""
        self._cpp_backend.forward(mu_x, var_x)

        return self.get_outputs()

    def backward(self):
        """Perform a backward pass."""
        self._cpp_backend.backward()

    def smoother(self):
        """Perform a smoother pass."""
        self._cpp_backend.smoother()

        return self.get_outputs_smoother()

    def step(self):
        """Perform a step of inference."""
        self._cpp_backend.step()

    def reset_lstm_states(self):
        """Reset lstm states"""
        self._cpp_backend.reset_lstm_states()

    def output_to_host(self) -> List[float]:
        """Copy the output data to the host."""
        return self._cpp_backend.output_to_host()

    def delta_z_to_host(self) -> List[float]:
        """Copy the delta Z data to the host."""
        return self._cpp_backend.delta_z_to_host()

    def set_delta_z(self, delta_mu: np.ndarray, delta_var: np.ndarray):
        """Send the delta Z to device"""
        self._cpp_backend.set_delta_z(delta_mu, delta_var)

    def get_layer_stack_info(self) -> str:
        """Get information about the layer stack."""
        return self._cpp_backend.get_layer_stack_info()

    def preinit_layer(self):
        """Preinitialize the layer."""
        self._cpp_backend.preinit_layer()

    def get_neg_var_w_counter(self) -> dict:
        """Get the number of negative variance weights."""
        return self._cpp_backend.get_neg_var_w_counter()

    def save(self, filename: str):
        """Save the model to a file."""
        self._cpp_backend.save(filename)

    def load(self, filename: str):
        """Load the model from a file."""
        self._cpp_backend.load(filename)

    def save_csv(self, filename: str):
        """Save the model parameters to a CSV file."""
        self._cpp_backend.save_csv(filename)

    def load_csv(self, filename: str):
        """Load the model parameters from a CSV file."""
        self._cpp_backend.load_csv(filename)

    def parameters(self) -> List[np.ndarray]:
        """Get the model parameters. Stored tuple (mu_w, var_w, mu_b, var_b) in a list"""
        return self._cpp_backend.parameters()

    def load_state_dict(self, state_dict: dict):
        """Load the model parameters from a state dict."""
        self._cpp_backend.load_state_dict(state_dict)

    def state_dict(self) -> dict:
        """Get the model parameters as a state dict where key is the layer name
        and value is a tuple of 4 arrays (mu_w, var_w, mu_b, var_b)"""
        return self._cpp_backend.state_dict()

    def params_from(self, other: "Sequential"):
        """Copy parameters from another model."""
        self._cpp_backend.params_from(other)

    def get_outputs(self) -> Tuple[np.ndarray, np.ndarray]:
        return self._cpp_backend.get_outputs()

    def get_outputs_smoother(self) -> Tuple[np.ndarray, np.ndarray]:
        return self._cpp_backend.get_outputs_smoother()

    def get_input_states(self) -> Tuple[np.ndarray, np.ndarray]:
        """Get the input states."""
        return self._cpp_backend.get_input_states()

    def get_norm_mean_var(self) -> dict:
        """Get the mean and variance of the normalization layer.
        Returns:
            A dictionary containing the mean and variance of the normalization layer.
            each key is the layer name and the value is a tuple of 4 arrays:
            mu_batch: mean of the batch
            var_batch: variance of the batch
            mu_ema_batch: mean of the exponential moving average (ema) of the batch
            var_ema_batch: variance of the ema of the batch

        """
        return self._cpp_backend.get_norm_mean_var()

    def get_lstm_states(self) -> dict:
        """
        Get the LSTM states for all LSTM layers as a dictionary.

        Returns:
            dict: A dictionary where each key is the layer index (int) and each value is a 4-tuple
                of numpy arrays (mu_h_prior, var_h_prior, mu_c_prior, var_c_prior).
        """
        return self._cpp_backend.get_lstm_states()

    def set_lstm_states(self, states: dict) -> None:
        """
        Set the LSTM states for all LSTM layers using a dictionary.

        Args:
            states (dict): A dictionary mapping layer indices (int) to a 4-tuple of numpy arrays:
                        (mu_h_prior, var_h_prior, mu_c_prior, var_c_prior).
        """
        self._cpp_backend.set_lstm_states(states)



import cutagi

from pytagi.nn.base_layer import BaseLayer


class ReLU(BaseLayer):
    """ReLU"""

    def __init__(self):
        self._cpp_backend = cutagi.ReLU()

    def get_layer_info(self) -> str:
        return self._cpp_backend.get_layer_info()

    def get_layer_name(self) -> str:
        return self._cpp_backend.get_layer_name()


class Sigmoid(BaseLayer):
    """Sigmoid"""

    def __init__(self):
        self._cpp_backend = cutagi.Sigmoid()

    def get_layer_info(self) -> str:
        return self._cpp_backend.get_layer_info()

    def get_layer_name(self) -> str:
        return self._cpp_backend.get_layer_name()


class Tanh(BaseLayer):
    """Tanh"""

    def __init__(self):
        self._cpp_backend = cutagi.Tanh()

    def get_layer_info(self) -> str:
        return self._cpp_backend.get_layer_info()

    def get_layer_name(self) -> str:
        return self._cpp_backend.get_layer_name()


class MixtureReLU(BaseLayer):
    """Mixture ReLU"""

    def __init__(self):
        self._cpp_backend = cutagi.MixtureReLU()

    def get_layer_info(self) -> str:
        return self._cpp_backend.get_layer_info()

    def get_layer_name(self) -> str:
        return self._cpp_backend.get_layer_name()


class MixtureSigmoid(BaseLayer):
    """Mixture Sigmoid"""

    def __init__(self):
        self._cpp_backend = cutagi.MixtureSigmoid()

    def get_layer_info(self) -> str:
        return self._cpp_backend.get_layer_info()

    def get_layer_name(self) -> str:
        return self._cpp_backend.get_layer_name()


class MixtureTanh(BaseLayer):
    """Mixture Tanh"""

    def __init__(self):
        self._cpp_backend = cutagi.MixtureTanh()

    def get_layer_info(self) -> str:
        return self._cpp_backend.get_layer_info()

    def get_layer_name(self) -> str:
        return self._cpp_backend.get_layer_name()


class Softplus(BaseLayer):
    """Softplus"""

    def __init__(self):
        self._cpp_backend = cutagi.Softplus()

    def get_layer_info(self) -> str:
        return self._cpp_backend.get_layer_info()

    def get_layer_name(self) -> str:
        return self._cpp_backend.get_layer_name()


class LeakyReLU(BaseLayer):
    """Leaky ReLU"""

    def __init__(self):
        self._cpp_backend = cutagi.LeakyReLU()

    def get_layer_info(self) -> str:
        return self._cpp_backend.get_layer_info()

    def get_layer_name(self) -> str:
        return self._cpp_backend.get_layer_name()


class Softmax(BaseLayer):
    """Softmax"""

    def __init__(self):
        self._cpp_backend = cutagi.Softmax()

    def get_layer_info(self) -> str:
        return self._cpp_backend.get_layer_info()

    def get_layer_name(self) -> str:
        return self._cpp_backend.get_layer_name()


class EvenExp(BaseLayer):
    """EvenExp"""

    def __init__(self):
        self._cpp_backend = cutagi.EvenExp()

    def get_layer_info(self) -> str:
        return self._cpp_backend.get_layer_info()

    def get_layer_name(self) -> str:
        return self._cpp_backend.get_layer_name()



import numpy as np

from pytagi.tagi_utils import HRCSoftmax, Utils


class HRCSoftmaxMetric:
    """Classifcation error for hierarchical softmax"""

    def __init__(self, num_classes: int):
        self.num_classes = num_classes
        self.utils = Utils()
        self.hrc_softmax: HRCSoftmax = self.utils.get_hierarchical_softmax(
            num_classes=num_classes
        )

    def error_rate(
        self, m_pred: np.ndarray, v_pred: np.ndarray, label: np.ndarray
    ) -> float:
        """Compute error rate for classifier"""
        batch_size = m_pred.shape[0] // self.hrc_softmax.len
        pred, _ = self.utils.get_labels(
            m_pred, v_pred, self.hrc_softmax, self.num_classes, batch_size
        )
        return classification_error(pred, label)

    def get_predicted_labels(
        self, m_pred: np.ndarray, v_pred: np.ndarray
    ) -> np.ndarray:
        """Get the prediction"""
        batch_size = m_pred.shape[0] // self.hrc_softmax.len
        pred, _ = self.utils.get_labels(
            m_pred, v_pred, self.hrc_softmax, self.num_classes, batch_size
        )
        return pred


def mse(prediction: np.ndarray, observation: np.ndarray) -> float:
    """Mean squared error"""
    return np.nanmean((prediction - observation) ** 2)


def log_likelihood(
    prediction: np.ndarray, observation: np.ndarray, std: np.ndarray
) -> float:
    """Compute the averaged log-likelihood"""

    log_lik = -0.5 * np.log(2 * np.pi * (std**2)) - 0.5 * (
        ((observation - prediction) / std) ** 2
    )

    return np.nanmean(log_lik)


def rmse(prediction: np.ndarray, observation: np.ndarray) -> None:
    """Root mean squared error"""
    mse = mse(prediction, observation)

    return mse**0.5


def classification_error(prediction: np.ndarray, label: np.ndarray) -> None:
    """Compute the classification error"""
    count = 0
    for pred, lab in zip(prediction.T, label):
        if pred != lab:
            count += 1

    return count / len(prediction)



from typing import Union

import cutagi

from pytagi.nn.base_layer import BaseLayer
from pytagi.nn.data_struct import BaseDeltaStates, BaseHiddenStates
from pytagi.nn.layer_block import LayerBlock


class ResNetBlock(BaseLayer):
    """A residual architecture contains a main block and a shortcut layer"""

    def __init__(
        self,
        main_block: Union[BaseLayer, LayerBlock],
        shortcut: Union[BaseLayer, LayerBlock] = None,
    ):
        if shortcut is not None:
            self._cpp_backend = cutagi.ResNetBlock(
                main_block._cpp_backend, shortcut._cpp_backend
            )
        else:
            self._cpp_backend = cutagi.ResNetBlock(main_block._cpp_backend)

    def init_shortcut_state(self) -> None:
        """Initialize state buffer for shortcut"""
        self._cpp_backend.init_shortcut_state()

    def init_shortcut_delta_state(self) -> None:
        """Initialize update values for state buffer for the shortcut"""
        self._cpp_backend.init_shortcut_delta_state()

    def init_input_buffer(self) -> None:
        """Initialize input state buffer to hold temporary state"""
        self._cpp_backend.init_input_buffer()

    @property
    def main_block(self) -> LayerBlock:
        """Set main block"""
        return self._cpp_backend.main_block

    @main_block.setter
    def main_block(self, value: LayerBlock):
        """Set main block"""
        self._cpp_backend.main_block = value

    @property
    def shortcut(self) -> BaseLayer:
        """Set shortcut"""
        return self._cpp_backend.shortcut

    @shortcut.setter
    def shortcut(self, value: BaseLayer):
        """Set shortcut"""
        self._cpp_backend.shortcut = value

    @property
    def input_z(self) -> BaseHiddenStates:
        """Get output hidden states"""
        return self._cpp_backend.input_z

    @input_z.setter
    def input_z(self, value: BaseHiddenStates):
        """Set input hidden states."""
        self._cpp_backend.input_z = value

    @property
    def input_delta_z(self) -> BaseDeltaStates:
        """Get update values for input states"""
        return self._cpp_backend.input_delta_z

    @input_delta_z.setter
    def input_delta_z(self, value: BaseDeltaStates):
        """Set update values for input states"""
        self._cpp_backend.input_delta_z = value

    @property
    def shortcut_output_z(self) -> BaseHiddenStates:
        """Get output hidden states for shortcut"""
        return self._cpp_backend.shortcut_output_z

    @shortcut_output_z.setter
    def shortcut_output_z(self, value: BaseHiddenStates):
        """Set output hidden states for shortcut"""
        self._cpp_backend.shortcut_output_z = value

    @property
    def shortcut_output_delta_z(self) -> BaseDeltaStates:
        """Get update values for output hidden states for shortcut"""
        return self._cpp_backend.shortcut_output_delta_z

    @shortcut_output_delta_z.setter
    def shortcut_output_delta_z(self, value: BaseDeltaStates):
        """Set update values for output hidden states for shortcut"""
        self._cpp_backend.shortcut_output_delta_z = value



from pytagi.nn.activation import (
    EvenExp,
    LeakyReLU,
    MixtureReLU,
    MixtureSigmoid,
    MixtureTanh,
    ReLU,
    Sigmoid,
    Softmax,
    Softplus,
    Tanh,
)
from pytagi.nn.base_layer import BaseLayer
from pytagi.nn.batch_norm import BatchNorm2d
from pytagi.nn.conv2d import Conv2d
from pytagi.nn.convtranspose2d import ConvTranspose2d
from pytagi.nn.data_struct import BaseDeltaStates, BaseHiddenStates, HRCSoftmax
from pytagi.nn.ddp import DDPConfig, DDPSequential
from pytagi.nn.layer_block import LayerBlock
from pytagi.nn.layer_norm import LayerNorm
from pytagi.nn.linear import Linear
from pytagi.nn.lstm import LSTM
from pytagi.nn.output_updater import OutputUpdater
from pytagi.nn.pooling import AvgPool2d, MaxPool2d
from pytagi.nn.resnet_block import ResNetBlock
from pytagi.nn.sequential import Sequential
from pytagi.nn.slinear import SLinear
from pytagi.nn.slstm import SLSTM



from typing import Tuple, Union

import cutagi
import numpy as np

from pytagi.nn import HRCSoftmax


class Utils:
    """Frontend for utility functions from C++/CUDA backend

    Attributes:
        _cpp_backend: Utility functionalities from the backend
    """

    def __init__(self) -> None:
        self._cpp_backend = cutagi.Utils()

    def label_to_obs(
        self, labels: np.ndarray, num_classes: int
    ) -> Tuple[np.ndarray, np.ndarray, int]:
        """Get observations and observation indices of the binary tree for
            classification

        Args:
            labels: Labels of dataset
            num_classes: Total number of classes
        Returns:
            obs: Encoded observations of the labels
            obs_idx: Indices of the encoded observations in the output vector
            num_obs: Number of encoded observations
        """

        obs, obs_idx, num_obs = self._cpp_backend.label_to_obs_wrapper(
            labels, num_classes
        )

        return np.array(obs), np.array(obs_idx), np.array(num_obs)

    def label_to_one_hot(
        self, labels: np.ndarray, num_classes: int
    ) -> np.ndarray:
        """Get the one hot encoder for each class

        Args:
            labels: Labels of dataset
            num_classes: Total number of classes
        Returns:
            one_hot: One hot encoder
        """

        return self._cpp_backend.label_to_one_hot_wrapper(labels, num_classes)

    def load_mnist_images(
        self, image_file: str, label_file: str, num_images: int
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Load mnist dataset

        Args:
            image_file: Location of the Mnist image file
            label_file: Location of the Mnist label file
            num_images: Number of images to be loaded
        Returns:
            images: Image dataset
            labels: Label dataset
            num_images: Total number of images
        """
        images, labels = self._cpp_backend.load_mnist_dataset_wrapper(
            image_file, label_file, num_images
        )

        return images, labels

    def load_cifar_images(
        self, image_file: str, num: int
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Load cifar dataset

        Args:
            image_file: Location of image file
            num: Number of images to be loaded
        Returns:
            images: Image dataset
            labels: Label dataset
        """

        images, labels = self._cpp_backend.load_cifar_dataset_wrapper(
            image_file, num
        )

        return images, labels

    def get_labels(
        self,
        ma: np.ndarray,
        Sa: np.ndarray,
        hr_softmax: HRCSoftmax,
        num_classes: int,
        batch_size: int,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Convert last layer's hidden state to labels

        Args:
            ma: Mean of activation units for the output layer
            Sa: Variance of activation units for the output layer
            hr_softmax: Hierarchical softmax
            num_classes: Total number of classes
            batch_size: Number of data in a batch
        Returns:
            pred: Label prediciton
            prob: Probability for each label
        """

        pred, prob = self._cpp_backend.get_labels_wrapper(
            ma, Sa, hr_softmax, num_classes, batch_size
        )

        return pred, prob

    def get_errors(
        self,
        ma: np.ndarray,
        Sa: np.ndarray,
        labels: np.ndarray,
        hr_softmax: HRCSoftmax,
        num_classes: int,
        batch_size: int,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Convert last layer's hidden state to labels

        Args:
            ma: Mean of activation units for the output layer
            Sa: Variance of activation units for the output layer
            labels: Label dataset
            hr_softmax: Hierarchical softmax
            num_classes: Total number of classes
            batch_size: Number of data in a batch
        Returns:
            pred: Label prediction
            prob: Probability for each label
        """

        pred, prob = self._cpp_backend.get_error_wrapper(
            ma, Sa, labels, hr_softmax, num_classes, batch_size
        )

        return pred, prob

    def get_hierarchical_softmax(self, num_classes: int) -> HRCSoftmax:
        """Convert labels to binary tree

        Args:
            num_classes: Total number of classes
        Returns:
            hr_softmax: Hierarchical softmax
        """
        hr_softmax = self._cpp_backend.hierarchical_softmax_wrapper(num_classes)

        return hr_softmax

    def obs_to_label_prob(
        self,
        ma: np.ndarray,
        Sa: np.ndarray,
        hr_softmax: HRCSoftmax,
        num_classes: int,
    ) -> np.ndarray:
        """Convert observation to label probabilities

        Args:
            ma: Mean of activation units for the output layer
            Sa: Variance of activation units for the output layer
            hr_softmax: Hierarchical softmax
            num_classes: Total number of classes
        Returns:
            prob: Probability for each label
        """

        prob = self._cpp_backend.obs_to_label_prob_wrapper(
            ma, Sa, hr_softmax, num_classes
        )

        return np.array(prob)

    def create_rolling_window(
        self,
        data: np.ndarray,
        output_col: np.ndarray,
        input_seq_len: int,
        output_seq_len: int,
        num_features: int,
        stride: int,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Create rolling window for time series data

        Args:
            data: dataset
            output_col: Indices of the output columns
            input_seq_len: Length of the input sequence
            output_seq_len: Length of the output sequence
            num_features: Number of features
            stride: Controls number of steps for the window movements
        Returns:
            input_data: Input data for neural networks in sequence
            output_data: Output data for neural networks in sequence
        """
        num_data = int(
            # (len(data) / num_features - input_seq_len - output_seq_len) / stride + 1
            (len(data) - input_seq_len - output_seq_len) / stride
            + 1
        )

        input_data, output_data = (
            self._cpp_backend.create_rolling_window_wrapper(
                data.flatten(),
                output_col,
                input_seq_len,
                output_seq_len,
                num_features,
                stride,
            )
        )
        # input_data = input_data.reshape((num_data, input_seq_len))
        input_data = input_data.reshape(
            (num_data, input_seq_len * num_features)
        )
        output_data = output_data.reshape((num_data, output_seq_len))

        return input_data, output_data

    def get_upper_triu_cov(
        self, batch_size: int, num_data: int, sigma: float
    ) -> np.ndarray:
        """Create an upper triangle covriance matrix for inputs"""

        vx_f = self._cpp_backend.get_upper_triu_cov_wrapper(
            batch_size, num_data, sigma
        )

        return np.array(vx_f)


def exponential_scheduler(
    curr_v: float, min_v: float, decaying_factor: float, curr_iter: float
) -> float:
    """Exponentially decaying"""

    return np.maximum(curr_v * (decaying_factor**curr_iter), min_v)


class Normalizer:
    """Different method to normalize the data before feeding
    to neural networks"""

    def __init__(self, method: Union[str, None] = None) -> None:
        self.method = method

    @staticmethod
    def standardize(
        data: np.ndarray, mu: np.ndarray, std: np.ndarray
    ) -> np.ndarray:
        """Z-score normalization where
        data_norm = (data - data_mean) / data_std"""

        return (data - mu) / (std + 1e-10)

    @staticmethod
    def unstandardize(
        norm_data: np.ndarray, mu: np.ndarray, std: np.ndarray
    ) -> np.ndarray:
        """Transform standardized data to original space"""
        return norm_data * (std + 1e-10) + mu

    @staticmethod
    def unstandardize_std(norm_std: np.ndarray, std: np.ndarray) -> np.ndarray:
        """Transform standardized std to original space"""

        return norm_std * (std + 1e-10)

    def max_min_norm(
        self, data: np.ndarray, max_value: np.ndarray, min_value: np.ndarray
    ) -> np.ndarray:
        """Normalize the data between 0 and 1"""
        assert np.all(max_value > min_value)
        return (data - min_value) / (max_value - min_value + 1e-10)

    @staticmethod
    def max_min_unnorm(
        norm_data: np.ndarray, max_value: np.ndarray, min_value: np.ndarray
    ) -> np.ndarray:
        """Transform max-min normalized data to original space"""

        return (norm_data * (max_value - min_value + 1e-10)) + min_value

    @staticmethod
    def max_min_unnorm_std(
        norm_std: np.ndarray, max_value: np.ndarray, min_value: np.ndarray
    ) -> np.ndarray:
        """Transform max-min normalized std to original space"""

        return norm_std * (max_value - min_value + 1e-10)

    @staticmethod
    def compute_mean_std(data: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Compute sample mean and standard deviation"""

        return (np.nanmean(data, axis=0), np.nanstd(data, axis=0))

    @staticmethod
    def compute_max_min(data: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Compute max min values"""

        return (np.nanmax(data, axis=0), np.nanmin(data, axis=0))



import cutagi
import numpy as np

from pytagi.nn.data_struct import BaseDeltaStates, BaseHiddenStates


class OutputUpdater:
    def __init__(self, model_device: str):
        self._cpp_backend = cutagi.OutputUpdater(model_device)

    def update(
        self,
        output_states: BaseHiddenStates,
        mu_obs: np.ndarray,
        var_obs: np.ndarray,
        delta_states: BaseDeltaStates,
    ):
        self._cpp_backend.update(
            output_states, mu_obs.tolist(), var_obs.tolist(), delta_states
        )

    def update_using_indices(
        self,
        output_states: BaseHiddenStates,
        mu_obs: np.ndarray,
        var_obs: np.ndarray,
        selected_idx: np.ndarray,
        delta_states: BaseDeltaStates,
    ):
        self._cpp_backend.update_using_indices(
            output_states,
            mu_obs.tolist(),
            var_obs.tolist(),
            selected_idx.tolist(),
            delta_states,
        )

    def update_heteros(
        self,
        output_states: BaseHiddenStates,
        mu_obs: np.ndarray,
        delta_states: BaseDeltaStates,
    ):
        self._cpp_backend.update_heteros(
            output_states, mu_obs.tolist(), delta_states
        )

    @property
    def device(self) -> str:
        return self._cpp_backend.device

    @device.setter
    def device(self, value: str):
        self._cpp_backend.device = value



# cutagi/python_src/__version__.py
import os

ROOT_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__)))
with open(os.path.join(ROOT_DIR, "version.txt"), "r") as f:
    version = f.readlines()
VERSION = version[0]

__version__ = VERSION



import cutagi
import numpy as np


class BaseLayer:
    """Base layer"""

    def __init__(self):
        self._cpp_backend = cutagi.BaseLayer()

    def to_cuda(self):
        self._cpp_backend.to_cuda()

    def get_layer_info(self) -> str:
        return self._cpp_backend.get_layer_info()

    def get_layer_name(self) -> str:
        return self._cpp_backend.get_layer_name()

    def get_max_num_states(self) -> int:
        return self._cpp_backend.get_max_num_states()

    @property
    def input_size(self) -> int:
        return self._cpp_backend.input_size

    @input_size.setter
    def input_size(self, value: int):
        self._cpp_backend.input_size = value

    @property
    def output_size(self) -> int:
        return self._cpp_backend.output_size

    @output_size.setter
    def output_size(self, value: int):
        self._cpp_backend.output_size = value

    @property
    def in_width(self) -> int:
        return self._cpp_backend.in_width

    @in_width.setter
    def in_width(self, value: int):
        self._cpp_backend.in_width = value

    @property
    def in_height(self) -> int:
        return self._cpp_backend.in_height

    @in_height.setter
    def in_height(self, value: int):
        self._cpp_backend.in_height = value

    @property
    def in_channels(self) -> int:
        return self._cpp_backend.in_channels

    @in_channels.setter
    def in_channels(self, value: int):
        self._cpp_backend.in_channels = value

    @property
    def out_width(self) -> int:
        return self._cpp_backend.out_width

    @out_width.setter
    def out_width(self, value: int):
        self._cpp_backend.out_width = value

    @property
    def out_height(self) -> int:
        return self._cpp_backend.out_height

    @out_height.setter
    def out_height(self, value: int):
        self._cpp_backend.out_height = value

    @property
    def out_channels(self) -> int:
        return self._cpp_backend.out_channels

    @out_channels.setter
    def out_channels(self, value: int):
        self._cpp_backend.out_channels = value

    @property
    def bias(self) -> bool:
        return self._cpp_backend.bias

    @bias.setter
    def bias(self, value: bool):
        self._cpp_backend.bias = value

    @property
    def num_weights(self) -> int:
        return self._cpp_backend.num_weights

    @num_weights.setter
    def num_weights(self, value: int):
        self._cpp_backend.num_weights = value

    @property
    def num_biases(self) -> int:
        return self._cpp_backend.num_biases

    @num_biases.setter
    def num_biases(self, value: int):
        self._cpp_backend.num_biases = value

    @property
    def mu_w(self) -> np.ndarray:
        return self._cpp_backend.mu_w

    @mu_w.setter
    def mu_w(self, value: np.ndarray):
        self._cpp_backend.mu_w = value

    @property
    def var_w(self) -> np.ndarray:
        return self._cpp_backend.var_w

    @var_w.setter
    def var_w(self, value: np.ndarray):
        self._cpp_backend.var_w = value

    @property
    def mu_b(self) -> np.ndarray:
        return self._cpp_backend.mu_b

    @mu_b.setter
    def mu_b(self, value: np.ndarray):
        self._cpp_backend.mu_b = value

    @property
    def var_b(self) -> np.ndarray:
        return self._cpp_backend.var_b

    @var_b.setter
    def var_b(self, value: np.ndarray):
        self._cpp_backend.var_b = value

    @property
    def delta_mu_w(self) -> np.ndarray:
        return self._cpp_backend.delta_mu_w

    @delta_mu_w.setter
    def delta_mu_w(self, value: np.ndarray):
        self._cpp_backend.delta_mu_w = value

    @property
    def delta_var_w(self) -> np.ndarray:
        return self._cpp_backend.delta_var_w

    @delta_var_w.setter
    def delta_var_w(self, value: np.ndarray):
        self._cpp_backend.delta_var_w = value

    @property
    def delta_mu_b(self) -> np.ndarray:
        return self._cpp_backend.delta_mu_b

    @delta_mu_b.setter
    def delta_mu_b(self, value: np.ndarray):
        self._cpp_backend.delta_mu_b = value

    @property
    def delta_var_b(self) -> np.ndarray:
        return self._cpp_backend.delta_var_b

    @delta_var_b.setter
    def delta_var_b(self, value: np.ndarray):
        self._cpp_backend.delta_var_b = value

    @property
    def num_threads(self) -> int:
        return self._cpp_backend.num_threads

    @num_threads.setter
    def num_threads(self, value: int):
        self._cpp_backend.num_threads = value

    @property
    def training(self) -> bool:
        return self._cpp_backend.training

    @training.setter
    def training(self, value: bool):
        self._cpp_backend.training = value

    @property
    def device(self) -> bool:
        return self._cpp_backend.device

    @training.setter
    def device(self, value: str):
        self._cpp_backend.device = value



from typing import List

import cutagi

from pytagi.nn.base_layer import BaseLayer


class LayerNorm(BaseLayer):
    """Layer normalization"""

    def __init__(
        self, normalized_shape: List[int], eps: float = 1e-4, bias: bool = True
    ):
        self.normalized_shape = normalized_shape
        self.eps = eps
        self.is_bias = bias

        self._cpp_backend = cutagi.LayerNorm(normalized_shape, eps, bias)

    def get_layer_info(self) -> str:
        return self._cpp_backend.get_layer_info()

    def get_layer_name(self) -> str:
        return self._cpp_backend.get_layer_name()

    def init_weight_bias(self):
        self._cpp_backend.init_weight_bias()



from typing import List

import cutagi

from pytagi.nn.base_layer import BaseLayer


class LayerBlock(BaseLayer):
    """A stack of different layer derived from BaseLayer"""

    def __init__(self, *layers: BaseLayer):
        """
        Initialize the Sequential model with the given layers.
        Args:
            layers: A variable number of layers (instances of BaseLayer or derived classes).
        """
        backend_layers = [layer._cpp_backend for layer in layers]
        self._cpp_backend = cutagi.LayerBlock(backend_layers)

    def switch_to_cuda(self):
        """Convert all layers to cuda layer"""
        self._cpp_backend.switch_to_cuda()

    @property
    def layers(self) -> None:
        """Get layers"""
        return self._cpp_backend.layers

    @layers.setter
    def layers(self, value: List[BaseLayer]):
        """Set base layers"""
        self._cpp_backend.layers = value



from typing import List, Tuple

import cutagi
import numpy as np

from pytagi.nn.data_struct import BaseDeltaStates, BaseHiddenStates
from pytagi.nn.sequential import Sequential


class DDPConfig:
    """Configuration for distributed training"""

    def __init__(
        self,
        device_ids: List[int],
        backend: str = "nccl",
        rank: int = 0,
        world_size: int = 1,
    ):
        self._cpp_backend = cutagi.DDPConfig(
            device_ids, backend, rank, world_size
        )

    @property
    def device_ids(self) -> List[int]:
        return self._cpp_backend.device_ids

    @device_ids.setter
    def device_ids(self, value: List[int]):
        self._cpp_backend.device_ids = value

    @property
    def backend(self) -> str:
        return self._cpp_backend.backend

    @backend.setter
    def backend(self, value: str):
        self._cpp_backend.backend = value

    @property
    def rank(self) -> int:
        return self._cpp_backend.rank

    @rank.setter
    def rank(self, value: int):
        self._cpp_backend.rank = value

    @property
    def world_size(self) -> int:
        return self._cpp_backend.world_size

    @world_size.setter
    def world_size(self, value: int):
        self._cpp_backend.world_size = value


class DDPSequential:
    """Distributed training wrapper for Sequential models"""

    def __init__(
        self,
        model: Sequential,
        config: DDPConfig,
        average: bool = True,
    ):
        self._cpp_backend = cutagi.DDPSequential(
            model._cpp_backend, config._cpp_backend, average
        )
        self.model = model
        self.config = config

    @property
    def output_z_buffer(self) -> BaseHiddenStates:
        """Get the output hidden states"""
        return self.model.output_z_buffer

    @output_z_buffer.setter
    def output_z_buffer(self, value: BaseHiddenStates):
        """Set the output hidden states."""
        self.model.output_z_buffer = value

    @property
    def input_delta_z_buffer(self) -> BaseDeltaStates:
        """Get the delta hidden states"""
        return self.model.input_delta_z_buffer

    @input_delta_z_buffer.setter
    def input_delta_z_buffer(self, value: BaseDeltaStates):
        """Set the delta hidden states."""
        self.model.input_delta_z_buffer = value

    def __call__(
        self, mu_x: np.ndarray, var_x: np.ndarray = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Perform a forward pass"""
        return self.forward(mu_x, var_x)

    def forward(
        self, mu_x: np.ndarray, var_x: np.ndarray = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Perform a forward pass"""
        self._cpp_backend.forward(mu_x, var_x)
        return self._cpp_backend.get_outputs()

    def backward(self):
        """Perform a backward pass"""
        self._cpp_backend.backward()

    def step(self):
        """Perform a parameter update step"""
        self._cpp_backend.step()

    def train(self):
        """Set the model in training mode"""
        self._cpp_backend.train()

    def eval(self):
        """Set the model in evaluation mode"""
        self._cpp_backend.eval()

    def barrier(self):
        """Synchronize all processes"""
        self._cpp_backend.barrier()

    def get_outputs(self) -> Tuple[np.ndarray, np.ndarray]:
        """Get the outputs of the model"""
        return self._cpp_backend.get_outputs()

    def output_to_host(self):
        """Copy the output to the host"""
        self._cpp_backend.output_to_host()

    def get_device_with_index(self) -> str:
        """Get the device with index"""
        return self._cpp_backend.get_device_with_index()



import cutagi

from pytagi.nn.base_layer import BaseLayer


class LSTM(BaseLayer):
    """LSTM layer"""

    def __init__(
        self,
        input_size: int,
        output_size: int,
        seq_len: int,
        bias: bool = True,
        gain_weight: float = 1.0,
        gain_bias: float = 1.0,
        init_method: str = "He",
    ):
        super().__init__()
        self.input_size = input_size
        self.output_size = output_size
        self.seq_len = seq_len
        self.bias = bias
        self.gain_weight = gain_weight
        self.gain_bias = gain_bias
        self.init_method = init_method

        self._cpp_backend = cutagi.LSTM(
            input_size,
            output_size,
            seq_len,
            bias,
            gain_weight,
            gain_bias,
            init_method,
        )

    def get_layer_info(self) -> str:
        return self._cpp_backend.get_layer_info()

    def get_layer_name(self) -> str:
        return self._cpp_backend.get_layer_name()

    def init_weight_bias(self):
        self._cpp_backend.init_weight_bias()



import cutagi

from pytagi.nn.base_layer import BaseLayer


class Conv2d(BaseLayer):
    """Convolutional layer"""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        bias: bool = True,
        stride: int = 1,
        padding: int = 0,
        padding_type: int = 1,
        in_width: int = 0,
        in_height: int = 0,
        gain_weight: float = 1.0,
        gain_bias: float = 1.0,
        init_method: str = "He",
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.is_bias = bias
        self.stride = stride
        self.padding = padding
        self.padding_type = padding_type
        self.in_width = in_width
        self.in_height = in_height
        self.gain_weight = gain_weight
        self.gain_bias = gain_bias
        self.init_method = init_method

        self._cpp_backend = cutagi.Conv2d(
            in_channels,
            out_channels,
            kernel_size,
            bias,
            stride,
            padding,
            padding_type,
            in_width,
            in_height,
            gain_weight,
            gain_bias,
            init_method,
        )

    def get_layer_info(self) -> str:
        return self._cpp_backend.get_layer_info()

    def get_layer_name(self) -> str:
        return self._cpp_backend.get_layer_name()

    def init_weight_bias(self):
        self._cpp_backend.init_weight_bias()
